import 'dart:io';
import 'dart:convert';
import 'package:flutter/material.dart';

import 'package:dvote/dvote.dart';
import 'package:flutter_dotenv/flutter_dotenv.dart';

class VoteScreen extends StatefulWidget {
  @override
  _VoteScreenState createState() => _VoteScreenState();
}

class _VoteScreenState extends State<VoteScreen> {
  ProcessMetadata _processMeta;
  Map<String, dynamic> _pollVoteEnvelope;
  String _merkleProof;
  bool _envelopeSent;
  int _censusSize, _envelopeHeight, _remainingSeconds;
  bool _loading = true;
  String _error;

  @override
  void initState() {
    super.initState();
    initPlatformState();
  }

  // Platform messages are asynchronous, so we initialize in an async method.
  Future<void> initPlatformState() async {
    EntityMetadata entityMeta;
    ProcessMetadata processMeta;
    Map<String, dynamic> pollVoteEnvelope;
    String merkleProof;
    int censusSize, envelopeHeight, remainingSeconds;
    bool envelopeSent;
    String error;

    await DotEnv().load('.env');

    final networkId = DotEnv().env['NETWORK_ID'];
    final entityId = DotEnv().env['ENTITY_ID'];
    final entityEntryPoint = DotEnv().env["ENTITY_ENTRY_POINT"];

    EntityReference entity = EntityReference();
    entity.entityId = entityId;
    entity.entryPoints.addAll([entityEntryPoint]);

    GatewayInfo gwInfo = await getRandomDefaultGatewayInfo(networkId);
    final DVoteGateway dvoteGw =
        DVoteGateway(gwInfo.dvote, publicKey: gwInfo.publicKey);

    final Web3Gateway web3Gw = Web3Gateway(gwInfo.web3);

    try {
      entityMeta = await fetchEntity(entity, dvoteGw, web3Gw);
    } on SocketException {
      error = "Connecting to the Gateway failed";
    } catch (err) {
      if (err is String) {
        error = err;
        print(err);
      } else {
        error = "Unable to retrieve the entity metadata";
      }
    }

    // If the widget was removed from the tree while the asynchronous platform
    // message was in flight, we want to discard the reply rather than calling
    // setState to update our non-existent appearance.
    if (!mounted) return;

    if (error != null) {
      setState(() {
        _error = error;
        _loading = false;
      });
      return;
    }

    try {
      final pid = entityMeta.votingProcesses?.active[0];
      if (pid != null) {
        final String privateKey = await mnemonicToPrivateKey(
            DotEnv().env['MNEMONIC'],
            hdPath: DotEnv().env['PATH']);
        // final String pubKey = await mnemonicToPublicKey(mnemonic, hdPath: path);
        // final String pubKeyClaim = await digestHexClaim(pubKey);
        final String pubKeyClaim = DotEnv().env['CENSUS_PUB_KEY_CLAIM'];
        final String censusMerkleRoot = DotEnv().env['CENSUS_MERKLE_ROOT'];

        processMeta = await getProcessMetadata(pid, dvoteGw, web3Gw);

        merkleProof =
            await generateProof(censusMerkleRoot, pubKeyClaim, dvoteGw);

        pollVoteEnvelope =
            await generatePollVoteEnvelope([1], merkleProof, pid, privateKey);
        envelopeSent = await sumbitEnvelope(pollVoteEnvelope, pid, dvoteGw);
        censusSize = await getCensusSize(processMeta.census.merkleRoot, dvoteGw);
        envelopeHeight = await getEnvelopeHeight(pid, dvoteGw);
        remainingSeconds =
            await getProcessRemainingTime(pid, 1000, 2000, dvoteGw);
      }
    } on SocketException {
      error = "Connecting to the Gateway failed";
    } catch (err) {
      if (err is String) {
        error = err;
        print(err);
      } else {
        error = "Unable to retrieve the acive processes";
      }
    }

    setState(() {
      _error = error;
      _loading = false;
      _processMeta = processMeta;

      _pollVoteEnvelope = pollVoteEnvelope;
      _envelopeSent = envelopeSent;
      _merkleProof = merkleProof;
      _remainingSeconds = remainingSeconds;
      _envelopeHeight = envelopeHeight;
      _censusSize = censusSize;
    });
  }

  @override
  Widget build(BuildContext context) {
    if (_loading || _processMeta == null) {
      return Scaffold(
        appBar: AppBar(
          title: const Text('Voting'),
        ),
        body: Container(
          child: Text("Loading..."),
        ),
      );
    } else if (_error != null) {
      return Scaffold(
        appBar: AppBar(
          title: const Text('Voting'),
        ),
        body: Container(
          child: Text("Error: " + _error),
        ),
      );
    } else if (_processMeta == null) {
      return Scaffold(
        appBar: AppBar(
          title: const Text('Voting'),
        ),
        body: Container(
          child: Text("The voting process metadata is empty"),
        ),
      );
    }

    return Scaffold(
      appBar: AppBar(
        title: const Text('Voting'),
      ),
      body: ListView(
        children: <Widget>[
          Padding(
            padding: EdgeInsets.all(16),
            child: Column(
              children: <Widget>[
                Text("(WORK IN PROGRESS)"),
                Text(_processMeta.details.title["default"] ?? "(No name)"),
                Divider(),
                Text("POLL VOTE PACKAGE"),
                Text(_pollVoteEnvelope != null
                    ? jsonEncode(_pollVoteEnvelope)
                    : ''),
                Divider(),
                Text("VOTE RESPONSE FROM GW"),
                //Text(_envelopeSent != null ? _envelopeSent.toString() : '')
              ],
            ),
          ),
        ],
      ),
    );
  }
}
