import 'dart:convert';
import 'dart:math';
import 'package:convert/convert.dart';
import 'package:dvote/dvote.dart';
import 'package:dvote/util/parsers.dart';

import '../net/gateway.dart';
import "./entity.dart";
import "../blockchain/index.dart";
import '../models/dart/entity.pb.dart';
import '../models/dart/process.pb.dart';
import '../util/json-sign.dart';

var _random = Random.secure();

/// Fetch both the active and ended voting processes of an Entity
Future<List<ProcessMetadata>> fetchAllProcesses(
    EntityReference entityRef, DVoteGateway dvoteGw, Web3Gateway web3Gw) async {
  try {
    final entity = await fetchEntity(entityRef, dvoteGw, web3Gw);

    final List<String> processes = entity.votingProcesses?.active ?? [];
    processes.addAll(entity.votingProcesses?.ended ?? []);

    return getProcessesMetadata(processes, dvoteGw, web3Gw);
  } catch (err) {
    throw Exception("The voting processes can't be retrieved");
  }
}

/// Fetch the active voting processes of an Entity
Future<List<ProcessMetadata>> fetchActiveProcesses(
    EntityReference entityRef, DVoteGateway dvoteGw, Web3Gateway web3Gw) async {
  try {
    final entity = await fetchEntity(entityRef, dvoteGw, web3Gw);

    final List<String> processes = entity.votingProcesses?.active ?? [];

    return getProcessesMetadata(processes, dvoteGw, web3Gw);
  } catch (err) {
    throw Exception("The active voting processes can't be retrieved");
  }
}

/// Fetch the ended voting processes of an Entity
Future<List<ProcessMetadata>> fetchEndedProcesses(
    EntityReference entityRef, DVoteGateway dvoteGw, Web3Gateway web3Gw) async {
  try {
    final entity = await fetchEntity(entityRef, dvoteGw, web3Gw);

    final List<String> processes = entity.votingProcesses?.ended ?? [];

    return getProcessesMetadata(processes, dvoteGw, web3Gw);
  } catch (err) {
    throw Exception("The active voting processes can't be retrieved");
  }
}

/// Fetch the metadata for the given Process ID
Future<ProcessMetadata> getProcessMetadata(
    String processId, DVoteGateway dvoteGw, Web3Gateway web3Gw) {
  return getProcessesMetadata([processId], dvoteGw, web3Gw).then((result) {
    if (result is List && result.length > 0)
      return result[0];
    else
      return null;
  });
}

/// Fetch the metadata for the given Process ID's
Future<List<ProcessMetadata>> getProcessesMetadata(
    List<String> processIds, DVoteGateway dvoteGw, Web3Gateway web3Gw) {
  return Future.wait(processIds.map((processId) async {
    try {
      final pid = hex.decode(processId.substring(2));
      final processData =
          await callVotingProcessMethod(web3Gw.rpcUri, "get", [pid]);

      if (!(processData is List) || !(processData[1] is String))
        return null;
      else if (processData[5] is bool && processData[5] == true) return null;

      final String strMetadata =
          await fetchFileString(ContentURI(processData[1]), dvoteGw);
      return parseProcessMetadata(strMetadata);
    } catch (err) {
      return null;
    }
  })).then((result) => result.whereType<ProcessMetadata>().toList());
}

/// Submit vote Envelope to the gateway
Future<bool> sumbitEnvelope(Map<String, dynamic> voteEnvelope, String processId,
    DVoteGateway dvoteGw) async {
  if (voteEnvelope == null || processId == null || dvoteGw == null)
    throw Exception("Invalid parameters");
  try {
    // Encrypt vote envelope with the public key of the Relay
    Map<String, dynamic> reqParams = {
      "method": "submitEnvelope",
      "processId": processId,
      "payload": voteEnvelope,
    };
    Map<String, dynamic> response = await dvoteGw.request(reqParams);
    if (!(response is Map) || !(response["ok"] is bool)) {
      throw Exception("Invalid response received from the gateway");
    }
    return response["ok"];
  } catch (err) {
    throw Exception("The vote envelope could not be submitted");
  }
}

Future<String> generateSnarkVoteEnvelope(
    List<int> votes, String proof, String processId, String privateKey,
    {String hdPath}) async {
  throw Exception("unimplemented");
  // TODO: Generate hash of private key for nullifier as in Snarks
  /*
  String votePackage = _generateSnarkVotePackage(votes);
  Map<String, String>  package = {
    "processId": processId,
    "proof": proof,  
    "nullifier": "0x1234...",   
    "vote-package": votePackage,  
  };
  return jsonEncode(package);
  */
}

Future<Map<String, String>> generatePollVoteEnvelope(
    List<int> votes, String merkleProof, String processId, String privateKey,
    {String hdPath}) async {
  if (votes == null ||
      merkleProof == null ||
      processId == null ||
      privateKey == null) throw Exception("Invalid parameters");
  try {
    // TODO: Generate hash of private key for nullifier as in Snarks
    final nonce = _generateRandomNumber(32);

    String votePackage = _generatePollVotePackage(votes);
    Map<String, String> package = {
      "processId": processId,
      "proof": merkleProof,
      "nonce": nonce,
      "vote-package": votePackage
    };

    final signature = await signJsonPayload(package, privateKey);
    package["signature"] = signature;

    return package;
  } catch (error) {
    throw Exception("Poll vote Envelope could not be generated");
  }
}

/// Returns number of existing envelopes in the process
Future<int> getEnvelopeHeight(String processId, DVoteGateway dvoteGw) async {
  if (processId == null || dvoteGw == null)
    throw Exception("Invalid parameters");
  try {
    Map<String, dynamic> reqParams = {
      "method": "getEnvelopeHeight",
      "processId": processId,
    };
    Map<String, dynamic> response = await dvoteGw.request(reqParams);
    if (!(response is Map) || !(response["height"] is int)) {
      throw Exception("Invalid response received from the gateway");
    }
    return response["height"];
  } catch (err) {
    throw Exception("The envelope height could not be retrieved");
  }
}

/// Returns estimated process remaining time in seconds
Future<int> getProcessRemainingTime(String processId, int startBlock,
    int numberOfBlocks, DVoteGateway dvoteGw) async {
  if (processId == null || dvoteGw == null)
    throw Exception("Invalid parameters");
  try {
    int height = await getEnvelopeHeight(processId, dvoteGw);
    // TODO: Get from gw-dvote VoteAPI the average block time
    int averageBlockTime = 3600; //seconds
    int remaininBlocks = ((numberOfBlocks - startBlock) - height);
    if (remaininBlocks <= 0)
      return 0;
    else
      return remaininBlocks * averageBlockTime;
  } catch (err) {
    throw Exception("The process deadline could not be determined");
  }
}

// ////////////////////////////////////////////////////////////////////////////
// / Internal helpers
// ////////////////////////////////////////////////////////////////////////////

String _generateSnarkVotePackage(List<int> votes, String publicKey) {
  final nonce = _generateRandomNumber(32);

  // TODO: ENCRYPT IT WITH publicKey

  Map<String, dynamic> package = {
    "type": "snark-vote",
    "nonce": nonce,
    "votes": votes
  };
  return base64.encode(utf8.encode(jsonEncode(package)));
}

String _generatePollVotePackage(List<int> votes) {
  final nonce = _generateRandomNumber(32);
  Map<String, dynamic> package = {
    "type": "poll-vote",
    "nonce": nonce,
    "votes": votes
  };
  return base64.encode(utf8.encode(jsonEncode(package)));
}

String _generateRandomNumber(int digits) {
  var result = "";
  for (var i = 0; i < 6; i++) {
    result = result + _random.nextInt(9).toString();
  }
  return result;
}
